module Deus.FunctionalTexture where

import Graphics.Rendering.OpenGL

import Data.Array
import Data.Maybe

import Control.Monad
import qualified Control.Monad.State as State
import Random

import Deus.Noise

data InterpolationMode = IMLinear | IMCubic deriving (Show,Eq,Enum)

chooseInterp IMLinear = lerp
chooseInterp IMCubic = cerp

data FunctionalTexture = Gradient ColorMap
                       | Noise StdGen Int Int ColorMap
                       | Scale GLdouble FunctionalTexture
                       | Turbulence StdGen Int Int FunctionalTexture
                       | PTBlend FunctionalTexture FunctionalTexture
                         deriving (Show)
showPT (Gradient cm) = "Gradient"
showPT (Noise gen x n cm) = "Noise"
showPT (Scale s tex) = "Scale ("++showPT tex++")"
showPT (Turbulence gen x n tex) = "Turbulence ("++showPT tex++")"
showPT (PTBlend tex tex') = "PTBlend ("++showPT tex++") ("++showPT tex'++")"

--instance Show FunctionalTexture where show = showPT

-- Context (or 'zipper') for navigating around PTs
data PTContext = PTCScale GLdouble PTContext
               | PTCTurbulence StdGen Int Int PTContext
               | PTCBlendLeft FunctionalTexture PTContext
               | PTCBlendRight FunctionalTexture PTContext
               | PTCTop
                 deriving (Show)
-- Current location cursor is a subtree and a context
data PTCursor = PTC PTContext FunctionalTexture
                 deriving (Show)

-- arbitrary choice that 'down' means go down left leg
ptc_down' (PTC c (Scale s pt))           = Just$PTC (PTCScale s c) pt
ptc_down' (PTC c (Turbulence g x n pt))  = Just$PTC (PTCTurbulence g x n c) pt
ptc_down' (PTC c (PTBlend l r))          = Just$PTC (PTCBlendRight r c) l
ptc_down' _                              = Nothing
ptc_down ptc = fromMaybe ptc (ptc_down' ptc)

ptc_up' (PTC (PTCScale s c) pt)          = Just$PTC c (Scale s pt)
ptc_up' (PTC (PTCTurbulence g x n c) pt) = Just$PTC c (Turbulence g x n pt)
ptc_up' (PTC (PTCBlendLeft l c) pt)      = Just$PTC c (PTBlend l pt)
ptc_up' (PTC (PTCBlendRight r c) pt)     = Just$PTC c (PTBlend pt r)
ptc_up' _                                = Nothing
ptc_up ptc = fromMaybe ptc (ptc_up' ptc)

ptc_left' (PTC (PTCBlendLeft l c) pt)    = Just$PTC (PTCBlendRight pt c) l
ptc_left' _                              = Nothing
ptc_left ptc = fromMaybe ptc (ptc_left' ptc)

ptc_right' (PTC (PTCBlendRight r c) pt)  = Just$PTC (PTCBlendLeft pt c) r
ptc_right' _                             = Nothing
ptc_right ptc = fromMaybe ptc (ptc_right' ptc)

reconstruct (PTC PTCTop pt) = pt
reconstruct ptc = reconstruct (ptc_up ptc)

ptc_delete (PTC (PTCScale s c)           pt) = PTC c pt
ptc_delete (PTC (PTCTurbulence g x n c)  pt) = PTC c pt
ptc_delete (PTC (PTCBlendLeft l c)       pt) = PTC c pt
ptc_delete (PTC (PTCBlendRight r c)      pt) = PTC c pt
ptc_delete ptc                               = ptc

ptc_ins_scale        (PTC c pt) = PTC c (Scale 1.0 pt)
ptc_ins_turbulence   (PTC c pt) = PTC c (Turbulence (mkStdGen 0) 1 1 pt)
ptc_ins_blend        (PTC c pt) = PTC c (PTBlend pt (Gradient bluesky))
ptc_replace_gradient (PTC c pt) = PTC c (Gradient bluesky)
ptc_replace_noise    (PTC c pt) = PTC c (Noise (mkStdGen 0) 1 1 bluesky)

gradient :: GLdouble -> GLdouble -> GLdouble
gradient x y = (sin (x * 2 * pi) + 1)/2

ptEval :: FunctionalTexture -> (GLdouble -> GLdouble -> Color4 GLdouble)
ptEval (Gradient cm) = \x y -> cmEval cm (gradient x y)
ptEval (Noise gen mnh mxh cm) =
    let n = multiNoise gen mnh mxh cerp 
    in \x y -> cmEval cm ((n x y + 1)/2)
ptEval (Scale s tex) = let e = ptEval tex in \x y -> e (x*s) (y*s)
ptEval (Turbulence gen mnh mxh tex) = 
    let off = multiNoise gen mnh mxh cerp 
        e   = ptEval tex
    in \x y -> e (x+off x y) (y+off x y)
ptEval (PTBlend t t') =
    let e = ptEval t
        e' = ptEval t'
    in \x y -> (let Color4 r1 g1 b1 a1 = e x y
                    Color4 r2 g2 b2 a2 = e' x y
                in Color4 (lerp r2 r1 a1) (lerp g2 g1 a1) (lerp b2 b1 a1) a2)



data ColorMap = ColorMap (Color4 GLdouble) [(GLdouble,Color4 GLdouble)] (Color4 GLdouble) deriving (Show)

data CMCursor = CMC [(GLdouble,Color4 GLdouble)] (GLdouble,Color4 GLdouble)
              [(GLdouble,Color4 GLdouble)]

cmc_make (ColorMap start mid end) = CMC [] (0,start) (mid ++ [(1,end)])

cmc_left (CMC []     c r) = CMC [] c r
cmc_left (CMC (l:ls) c r) = CMC ls l (c:r)

cmc_right (CMC l c []    ) = CMC l     c []
cmc_right (CMC l c (r:rs)) = CMC (c:l) r rs

cmc_move_left (CMC [] c r) = CMC [] c r
cmc_move_left (CMC ((t0,c0):ls) (t,c) r) 
    | (t-0.1) > t0  = CMC ((t0,c0):ls) (t-0.1,c) r
    | otherwise     = let t' = max 0 (t-0.1) in
                      CMC ls (t',c) ((t0,c0):r)

cmc_move_right (CMC l c []) = CMC l c []
cmc_move_right (CMC l (t,c) ((t0,c0):r)) 
    | (t+0.1) < t0  = CMC l (t+0.1,c) ((t0,c0):r)
    | otherwise     = let t' = min 1 (t+10.1) in
                      CMC ((t0,c0):l) (t',c) r

cmc_del (CMC l c []) = CMC l c []
cmc_del (CMC [] c r) = CMC [] c r
cmc_del (CMC (l:ls) c r) = CMC ls l r

cmc_dup (CMC l c r) = CMC (c:l) c r

cmc_reconstruct (CMC l c r) = let all       = reverse l ++ [c] ++ r
                                  (_,start) = head all
                                  (_,end)   = last all
                                  mid       = init . tail $ all
                              in ColorMap start mid end

-- Context for colormap, in texture

data CMContext = CMCGradient PTContext
               | CMCNoise StdGen Int Int PTContext

-- cursor for colormap in texture

data CMCTCursor = CMCT CMContext CMCursor

cmct_up (CMCT (CMCGradient c) cmc) = PTC c (Gradient (cmc_reconstruct cmc))
cmct_up (CMCT (CMCNoise g x n c) cmc) = PTC c (Noise g x n (cmc_reconstruct cmc))

cmct_reconstruct = reconstruct . cmct_up


bluesky = ColorMap (Color4 0 0 1 1) [(0.5,Color4 1 1 1 1)] (Color4 0.5 0.5 0.5 1)
marble = ColorMap (Color4 0 0 0 0) [] (Color4 0.4 0 0.8 1)

cmEval :: ColorMap -> GLdouble -> Color4 GLdouble
cmEval (ColorMap start mids end) t = cmEval_ ((0,start):(mids++[(1,end)])) t

cmEval_ ((t0,start):(t1,end):rest) t
    | t <= t1   = color_erp start end ((t-t0)/(t1-t0))
    | otherwise = cmEval_ ((t1,end):rest) t

color_erp :: (Color4 GLdouble) -> (Color4 GLdouble) -> GLdouble
          -> (Color4 GLdouble)
color_erp (Color4 r1 g1 b1 a1) (Color4 r2 g2 b2 a2) t =
    Color4 (lerp r1 r2 t) (lerp g1 g2 t) (lerp b1 b2 t) (lerp a1 a2 t)

colorBytes :: Color4 GLdouble -> Color4 GLubyte
colorBytes (Color4 r g b a) = Color4 (ub r) (ub g) (ub b) (ub a)
  where ub x = round (x*255)

