module Deus.Noise where

import Graphics.Rendering.OpenGL
import Graphics.UI.GLUT

import Control.Monad
import Control.Monad.Fix
import Control.Concurrent
import qualified Control.Monad.State as State
import Random

import Data.Array.Unboxed
import Data.Maybe


-- General 2D functions have type (GLdouble -> GLdouble -> a)

-- then we have sampling functions which take a 2D function, and
-- a grid, and produce samples

-- Simple RandomGen state monad

randomRState (mn,mx) = State.State (\g -> randomR (mn,mx) g)

noise2Dd :: StdGen ->
            (GLdouble -> GLdouble -> GLdouble -> GLdouble) ->
            GLsizei ->
            (GLdouble -> GLdouble -> GLdouble)
noise2Dd rng erp r =
    let --ran :: State.State StdGen GLdouble
        ran = randomRState (-1,1)
        ranxy = do (x,y) <- liftM2 (,) ran ran
                   return (x/sqrt (x*x+y*y),y/sqrt (x*x+y*y))
    in State.evalState (do 
      -- random gradients at all the corners
      grads <- listArray ((0,0),(r,r)) `fmap` replicateM (fromIntegral $ (r+1)*(r+1)) ranxy :: State.State StdGen (Array (GLsizei,GLsizei) (GLdouble,GLdouble))
      let grad_lookup = (\(x,y) -> grads ! (x `mod` r,y `mod` r))
      return $ \x y -> interp2D erp r grad_lookup x y
                       ) rng

interp2D :: (GLdouble -> GLdouble -> GLdouble -> GLdouble) -> 
            GLsizei ->
            ((GLsizei,GLsizei) -> (GLdouble,GLdouble)) ->
            GLdouble -> GLdouble ->
            GLdouble
interp2D erp r grads x y =
    let cellx = floor (x*fromIntegral r)
        celly = floor (y*fromIntegral r)
        x' = (x-(fromIntegral cellx/fromIntegral r)) * (fromIntegral r)
        y' = (y-(fromIntegral celly/fromIntegral r)) * (fromIntegral r)
        [(gAx,gAy),(gBx,gBy),(gCx,gCy),(gDx,gDy)] =
            map grads [(cellx,celly  ),(cellx+1,celly  ),
                          (cellx,celly+1),(cellx+1,celly+1)]
    in interp2D1Cell erp (gAx,gAy) (gBx,gBy) (gCx,gCy) (gDx,gDy) x' y'

interp2D1Cell :: (GLdouble -> GLdouble -> GLdouble -> GLdouble) ->
                 (GLdouble,GLdouble) -> (GLdouble,GLdouble) -> 
                 (GLdouble,GLdouble) -> (GLdouble,GLdouble) -> 
                 GLdouble -> GLdouble -> 
                 GLdouble
interp2D1Cell erp (gAx,gAy) (gBx,gBy) (gCx,gCy) (gDx,gDy) x y =
    -- four planes determined by those gradients
    let pA x y = (gAx * x     + gAy * y)
        pB x y = (gBx * (x-1) + gBy * y)
        pC x y = (gCx * x     + gCy * (y-1))
        pD x y = (gDx * (x-1) + gDy * (y-1))
    -- calculate interpolated function by lerp between four points
    in  cerp (cerp (pA x y) (pB x y) x)
            (cerp (pC x y) (pD x y) x)
            y

multiNoise rng minharm maxharm erp =
  let harms = map (2^) [minharm..maxharm]
      scales = map (2^) [1..(maxharm-minharm+1)]
      noises = map (noise2Dd rng erp) harms
      scalednoises = zipWith (\s f -> (\x y -> f x y / s)) scales noises
  in foldr1 (\f g -> (\x y -> f x y + g x y)) scalednoises

sample2D :: GLsizei -> GLsizei -> 
            (GLdouble -> GLdouble -> a) ->
            [a]
sample2D w h f = 
    [ f (fromIntegral x/fromIntegral (w)) (fromIntegral y/fromIntegral (h))
      | x <- [0..(w-1)], y <- [0..(h-1)] ]

-- linear interp
lerp :: GLdouble -> GLdouble -> GLdouble -> GLdouble
lerp a b t = b * t  + a * (1-t)

-- cubic interp
cerp :: GLdouble -> GLdouble -> GLdouble -> GLdouble
cerp a b t = lerp a b (t*t*(3-2*t))

integralerp erp a b t = round (erp (fromIntegral a) (fromIntegral b) t)

toBytes :: (GLdouble,GLdouble) ->
           (GLdouble -> GLdouble -> GLdouble) ->
           (GLdouble -> GLdouble -> GLubyte)
toBytes (mn,mx) f x y = floor ((f x y + mn)/(mx-mn)*255)

noiseBytes :: StdGen -> Int -> Int ->
           (GLdouble -> GLdouble -> GLdouble -> GLdouble) ->
           (GLdouble -> GLdouble -> GLubyte)
noiseBytes rng minharm maxharm erp =
    toBytes (-1/sqrt 2,1/sqrt 2) $ multiNoise rng minharm maxharm erp
