module Deus.OBJLoader where

import System.Exit

import Data.Maybe

import qualified Data.Map as M

import Graphics.Rendering.OpenGL

import Foreign(withArray,newArray,nullPtr)
import Foreign.Storable

data OBJData = MaterialLib String
             | ObjectName String
             | Vertex GLfloat GLfloat GLfloat
             | TexCoord GLfloat GLfloat GLfloat
             | Normal GLfloat GLfloat GLfloat
             | UseMaterial String
             | Ess
             | Face [(Int,Int,Int)]
               deriving (Show)

loadfile = concatMap loadline . lines

-- skip empty lines
loadline [] = []
-- skip comments
loadline ('#':_) = []

loadline s =
    let (w:ws) = words s 
    in case w of
         "mtllib" -> [MaterialLib (unwords ws)]
         "o" -> [ObjectName (unwords ws)]
         "v" -> let [x,y,z] = map read ws in [Vertex x y z]
         "vt" -> let [x,y,z] = map read ws in [TexCoord x y z]
         "vn"-> let [x,y,z] = map read ws in [Normal x y z]
         "usemtl" -> [UseMaterial (unwords ws)]
         "s" -> [Ess]
         "f" -> [Face (map parsetriple ws)]

parsetriple s = let (a,'/':rest) = break (=='/') s
                    (b,'/':c)    = break (=='/') rest
                in (read a, read b, read c)

getvertex (Vertex x y z) = Just (Vertex3 x y z)
getvertex _              = Nothing

getnormal (Normal x y z) = Just (Normal3 x y z)
getnormal _              = Nothing

gettriangle (Face l) | length l == 3 = Just l
gettriangle _                        = Nothing

getquad (Face l) | length l == 4 = Just l
getquad _                        = Nothing

type Vert = Vertex3 GLfloat
type Norm = Normal3 GLfloat

data Model = Model { quads :: [(Vert,Norm)],
                     tris  :: [(Vert,Norm)],
                     verts :: [Vert],
                     norms :: [Norm]
                   }
           deriving (Show)
                    
data VBOModel = VBOModel { vboVerts :: BufferObject,
                           vboNorms :: BufferObject,
                           vboQuads :: BufferObject,
                           vboSize  :: GLsizei }
                
loadOBJ s = 
    let objdata  = loadfile s
        vertices = catMaybes . map getvertex   $ objdata
        normals  = catMaybes . map getnormal   $ objdata
        qs       = catMaybes . map getquad     $ objdata
        ts       = catMaybes . map gettriangle $ objdata
    in
      Model { quads = concatMap 
                      (map (\(v,_,n) -> (vertices!!(v-1),normals!!(n-1)))) qs,
              tris  = concatMap 
                      (map (\(v,_,n) -> (vertices!!(v-1),normals!!(n-1)))) ts,
              verts = vertices,
              norms = normals
            }

drawModel m = do
  renderPrimitive Quads $ mapM_ (\(v,n) -> normal n >> vertex v) (quads m)
  renderPrimitive Triangles $ mapM_ (\(v,n) -> normal n >> vertex v) (tris m)

modelToVBO m = do
  putStrLn $ "modelToVBO stats : " ++ show (length $ verts m) ++ " vertices,"
    ++ show (length $ norms m) ++ " normals,"
    ++ show (length $ tris m) ++ " triangles,"
    ++ show (length $ quads m) ++ " quads."
    -- build up lists for vertices, normals and indices so that
    -- glDrawElements works - i.e. vertex and normal must have
    -- corresponding indices. Use a temporary map to cache old vertices
  let  (_,quadvs_r,quadns_r,quadis_r) = 
         foldr 
         (\(v,n) (vnmap,vs,ns,is) ->
           case M.lookup (v,n) vnmap of
             Just i -> -- this v/n pair already exists, just index it
               (vnmap,vs,ns,i:is)
             Nothing -> --this v/n pair doesn't exist, add it
               let newindex = length vs
               in (M.insert (v,n) newindex vnmap,
                   v:vs,
                   n:ns,
                   newindex:is))
         (M.empty,[],[],[])
         (quads m)
       quadvs = reverse quadvs_r
       quadns = reverse quadns_r
       quadis = reverse quadis_r
  putStrLn $ "Compacting complete, "
    ++ show (length $ quadvs) ++ " vertices, "
    ++ show (length $ quadns) ++ " normals, "
    ++ show (length $ quadis) ++ " indices."
  let newBufferObjectList buf xs mode = do
        [bo] <- genObjectNames 1
        bindBuffer buf $= Just bo
        withArray xs $ \p ->
          bufferData buf $= (fromIntegral (length xs * sizeOf (head xs)),p,mode)
        return bo
  vbo <- newBufferObjectList ArrayBuffer quadvs StaticDraw
  nbo <- newBufferObjectList ArrayBuffer quadns StaticDraw
  quadvbo <- newBufferObjectList ElementArrayBuffer quadis StaticDraw
  return (VBOModel { vboVerts = vbo,
                     vboNorms = nbo,
                     vboQuads = quadvbo,
                     vboSize  = fromIntegral (length quadis) })
  
drawVBO vbom = do
  clientState VertexArray $= Enabled
  clientState NormalArray $= Enabled
  clientState IndexArray $= Enabled
  bindBuffer ArrayBuffer $= Just (vboVerts vbom)
  arrayPointer VertexArray $= VertexArrayDescriptor 3 Float (fromIntegral $ sizeOf (undefined::Vert)) nullPtr
  bindBuffer ArrayBuffer $= Just (vboNorms vbom)
  arrayPointer NormalArray $= VertexArrayDescriptor 3 Float (fromIntegral $ sizeOf (undefined::Norm)) nullPtr
  bindBuffer ElementArrayBuffer $= Just (vboQuads vbom)
  drawElements Quads (vboSize vbom) UnsignedInt nullPtr
  clientState IndexArray $= Disabled
  clientState VertexArray $= Disabled
  clientState NormalArray $= Disabled
  