import System.IO

import Data.Monoid
import Data.Maybe
import Data.Char

import Control.Applicative
import Control.Monad
import Control.Monad.Reader
import Control.Monad.Instances

import Control.Concurrent

import Data.Time.Clock
import Data.Time.Calendar
import Data.Time.Format

import Deus.Reactive

  
-- -- demo1 : imperative approach
-- demo1run = do
--   c <- getChar
--   putStrLn $ "You typed " ++ show c
--   demo1run
  
--demo1 : reactive approach
demo1reactor :: Event Char -> Event (IO ())
demo1reactor = fmap (\c -> putStrLn $ "You typed " ++ show c) 

-- demo2 : composing three reactors
demo2reactorA :: Event Char -> Event (IO ())
demo2reactorA = fmap (\c -> 
                       do when (c == 'a') (putStrLn "That was a small a")
                          when (c == 'B') (putStrLn "That was a large B")
                     )
demo2reactorZ :: Event Char -> Event (IO ())
demo2reactorZ = fmap (\c -> putStrLn $ show c ++ " is a CAPITAL letter") .
                filterE isUpper
                

demo2reactorTick :: Event Int -> Event (IO ())
demo2reactorTick = fmap (\n -> putStrLn $ "TICK " ++ show n) 

demo2reactor kp tk = demo2reactorZ kp `mappend` 
                     demo2reactorA kp `mappend` 
                     demo2reactorTick tk
                     
-- demo3 : using a Reactive to keep track of last keypress

demo3reactive :: Event Char -> Reactive (Maybe Char)
demo3reactive kp = Nothing `Stepper` (Just `fmap` kp)

demo3reactor kp tk = snapshotWithR 
                     (\c t -> putStrLn $ "TICK " ++ show (t :: Int) ++ " last keypress was " ++ show c) (demo3reactive kp) tk

-- demo4 : Using a behaviour to track current time, and withTimeE
-- to annotate events with times, two ways to attach a time to a keypress event

demo4reactor :: Event Char -> Event (IO ())
demo4reactor kp = 
  snapshotWith
  (\t' (t,c) -> putStrLn $ "You pressed " ++ show c ++ " stamped at " 
  ++ show t ++ " snapshotted with " ++ show t')
  currentTime (withTimeE kp)

-- demo5 : A continously increasing counter. Press '<' or '>' to change speed
-- of increase. Value is shown every 0.8 seconds, approx

-- basic linearly increase behaviour, parametric on a start time
demo5behaviour :: Double -> UTCTime -> Double -> FunTime Double
demo5behaviour basevalue basetime speed = 
  FunTime (\t -> basevalue + realToFrac (t `diffUTCTime` basetime) * speed)
  -- timeFunction (\t -> basevalue + realToFrac (t `diffUTCTime` basetime) * speed)

-- speed changes are Events based on input chars
-- this version generates spurious events of '0 change' for other chars
demo5speedchange :: Event Char -> Event (Double -> Double)
demo5speedchange = fmap (\c -> case c of '<' -> (\x -> x-0.25)
                                         '>' -> (\x -> x+0.25)
                                         _   -> (\x -> x))

-- we build 'current speed' as a reactive XXXXXX [actually an event]
-- from an initial speed and the speed change Events
demo5speed :: Double -> Event Char -> Event Double
demo5speed init kp = accumE init (demo5speedchange kp)

--demo5value :: UTCTime -> Reactive (Behaviour Double)

-- just lacks initial values
demo5halfway :: UTCTime -> Double -> Event Char -> 
                Event (Double -> FunTime Double)
demo5halfway basetime init kp = 
  fmap (\(t,v) -> \s0 -> demo5behaviour s0 t v)
  (withTimeE (demo5speed init kp))

-- OK this one is a bit horrendous. The type checker helped me find it
-- but I'm not convinced I can explain it.
-- It's annoying that 'demo5behaviour' is FunTime and not Behaviour,
-- but that made this one quite a bit easier. 
demo5fullbehaviour t0 v0 kp = Behaviour $ O $
  demo5behaviour 0 t0 v0 `Stepper` 
  accumE (demo5behaviour 0 t0 v0) 
  (fmap ( \(t,f) -> (\(FunTime ft0) -> f (ft0 t) )) 
   (withTimeE (demo5halfway t0 v0 kp)))

-- and finally something to display some output
demo5reactor t0 kp ticker = 
  snapshotWith (\s n -> putStrLn $ "TICK : Current value is " ++ show s)
  (demo5fullbehaviour t0 1 kp)
  ticker
  --- these remaining two mappends are just debugging reactors!
  --- FRP means you can add "debugging printfs" externally. Sort of.
  `mappend`
  fmap (\v -> putStrLn $ "SPEED CHANGE : New speed is " ++ show v)
  (demo5speed 1 kp)
  `mappend` 
  fmap (\(t,ff) -> 
         let FunTime f = ff 0 in -- apply to initial value of 0 since we don't know 'real' initial value in demo5halfway
         putStrLn $ "Halfway: " ++ show t ++ "f(t) = " 
       ++ show (f t) ++ " f(t+1) - f(t) = " ++ show (f (addUTCTime 1 t) - f t))
  (withTimeE $ demo5halfway t0 1 kp)


-- demo 6 : a simple modal application
-- this program is two programs in one, switched between
-- using a menu screen. The two programs deal with input events
-- slightly differently.
-- The point is to demonstrate compositionality

-- program one adds up 1 digit numbers without printing anything
-- and prints the total when you type '='

demo6adderTotal :: Event Char -> Event Int
demo6adderTotal kp = 
  accumE 0 (fmap (\c -> case () of 
                     _ | c `elem` "0123456789" -> (+(digitToInt c))
                       | c == '='              -> const 0
                       | otherwise             -> id
                     ) kp)

demo6adderLastTotal :: Event Char -> Reactive Int
demo6adderLastTotal kp = 
  0 `Stepper` fmap (fromMaybe 0) (lastOne (demo6adderTotal kp))

demo6reactorAdderMode kp =  
  snapshotWithR 
  (\o c -> when (c=='=') . putStrLn $ "Total " ++ show o)
  (demo6adderLastTotal kp) (kp) 

demo6lines :: Event Char -> Event String
demo6lines kp = 
  catMaybesE . fmap snd $ 
  accumE ("",Nothing) (fmap (\c -> if c == '\n' 
                                   then (\ (l,_) -> ("",Just (reverse l))) 
                                   else (\ (l,_) -> (c:l,Nothing))) kp)

demo6reactorReverseMode kp = 
  fmap (\s -> putStrLn $ " backwards is " ++ reverse s) (demo6lines kp)

data Demo6Mode = MenuScreen
               | AdderMode
               | ReverseMode deriving (Show,Eq)
                 
demo6currentMode :: Event Char -> Reactive Demo6Mode
demo6currentMode kp = 
  accumR MenuScreen (fmap (\c -> \m -> case (m,c) of 
                              (MenuScreen,'1') -> AdderMode
                              (MenuScreen,'2') -> ReverseMode
                              (AdderMode,'q') -> MenuScreen
                              (ReverseMode,'\ESC') -> MenuScreen
                              _ -> m) kp)

demo6keysforadder kp =
  catMaybesE $ snapshotWithR (\m c -> case m of AdderMode -> Just c
                                                _ -> Nothing)
  (demo6currentMode kp) kp
  
demo6keysforreverse kp =
  catMaybesE $ snapshotWithR (\m c -> case m of ReverseMode -> Just c
                                                _ -> Nothing)
  (demo6currentMode kp) kp

demo6reactor kp = 
  mconcat
  [fmap (\s -> putStrLn $ "changing mode to " ++ show s) 
   (changesR $ demo6currentMode kp),
   demo6reactorReverseMode (demo6keysforreverse kp),
   demo6reactorAdderMode (demo6keysforadder kp)]  


------------------
--     Main     --
------------------

main = do
  hSetBuffering stdin NoBuffering
  hSetBuffering stdout NoBuffering
  keypresses <- getKeypresses
  ticker <- timewiseIterate 1000000 (+1) 0
  starttime <- getCurrentTime
  runReactor (demo5reactor starttime keypresses ticker)
